# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import re

import jinja2

from synalinks.src import ops
from synalinks.src.api_export import synalinks_export
from synalinks.src.backend import ChatMessage
from synalinks.src.backend import ChatMessages
from synalinks.src.backend import ChatRole
from synalinks.src.backend import Instructions
from synalinks.src.backend import Prediction
from synalinks.src.backend import SymbolicDataModel
from synalinks.src.backend import is_chat_messages
from synalinks.src.modules.module import Module
from synalinks.src.saving import serialization_lib

ROLES = [ChatRole.SYSTEM, ChatRole.USER, ChatRole.ASSISTANT, ChatRole.TOOL]

XML_TAGS_REGEX = re.compile(
    r"<(" + "|".join(ROLES) + r")\s*(?:[^>]*)>\s*([\s\S]*?)\s*</\1>",
    re.MULTILINE,
)


@synalinks_export("synalinks.default_prompt_template")
def default_prompt_template():
    """Returns the default prompt template.

    Returns:
        (str): The default prompt template.
    """
    return """
# Instructions
{{ instructions }}
{% if inputs_schema %}
# Input Schema
{{ inputs_schema }}
{% endif %}{% if outputs_schema %}
# Output schema
{{ outputs_schema }}
{% endif %}{% if examples %}
# Examples
{% for example in examples %}
## Input:
{{ example[0] }}
## Output:
{{ example[1] }}
{% endfor %}
{% endif %}
""".strip()


def default_instructions(data_model_fields):
    return f"""
Your task is to answer with a JSON containing the following keys: {data_model_fields}
""".strip()


@synalinks_export(["synalinks.modules.Generator", "synalinks.Generator"])
class Generator(Module):
    """
    Use a `LanguageModel` to generate a data model from an arbitrary input data model.

    Example:

    ```python
    import synalinks
    import asyncio

    async def main():

        class Query(DataModel):
            query: str = synalinks.Field(
                description="The user query",
            )

        class AnswerWithCritique(synalinks.DataModel):
            thinking: str = synalinks.Field(
                description="Your step by step thinking",
            )
            critique: str = synalinks.Field(
                description="The critique of the above thinking",
            )
            answer: str = synalinks.Field(
                description="The correct answer",
            )

        language_model = synalinks.LanguageModel(
            model="ollama/mistral",
        )

        x0 = synalinks.Input(data_model=Query)
        x1 = await synalinks.Generator(
            data_model=AnswerWithCritique,
            language_model=language_model,
        )(x0)

        program = synalinks.Program(
            inputs=x0,
            outputs=x1,
            name="chain_of_thought_with_critique",
            description="Useful to answer step by step and evaluate your answer",
        )

    if __name__ == "__main__":
        asyncio.run(main())
    ```

    Args:
        schema (dict): The target JSON schema.
            If not provided use the `data_model` to infer it.
        data_model (DataModel | SymbolicDataModel | JsonDataModel): The target data
            model for structured output.
        language_model (LanguageModel): The language model to use.
        prompt_template (str): The jinja2 prompt template.
        examples (list): The default list of examples, the examples
            are a list of tuples containing input/output JSON pairs.
        instructions (str): The default instructions being a string containing
            instructions for the language model.
        seed_instructions (list): Optional. A list of instructions to use as seed for the
            optimization. If not provided, use the default instructions as seed.
        use_inputs_schema (bool): Optional. Whether or not use the inputs schema in
            the prompt (Default to False).
        use_outputs_schema (bool): Optional. Whether or not use the outputs schema in
            the prompt (Default to False).
        return_inputs (bool): Optional. Whether or not to concatenate the inputs to
            the outputs (Default to False).
        temperature (float): Optional. The temperature for the LM call.
        streaming (str): Optional. If true stream the LM response, enabled only if
            `schema` is `None` and only during inference (not during training).
        name (str): Optional. The name of the module.
        description (str): Optional. The description of the module.
        trainable (bool): Whether the module's variables should be trainable.
    """

    def __init__(
        self,
        schema=None,
        data_model=None,
        language_model=None,
        prompt_template=None,
        examples=None,
        instructions=None,
        seed_instructions=None,
        use_inputs_schema=False,
        use_outputs_schema=False,
        return_inputs=False,
        temperature=0.0,
        streaming=False,
        name=None,
        description=None,
        trainable=True,
    ):
        super().__init__(
            name=name,
            description=description,
            trainable=trainable,
        )
        if not schema and data_model:
            schema = data_model.get_schema()
        self.schema = schema
        if not language_model:
            raise ValueError("You should provide `language_model` parameter.")
        self.language_model = language_model
        if not prompt_template:
            prompt_template = default_prompt_template()
        self.prompt_template = prompt_template
        if not examples:
            examples = []
        self.examples = examples
        if not instructions and self.schema:
            data_model_keys = list(self.schema["properties"].keys())
            instructions = default_instructions(data_model_keys)
        self.instructions = instructions
        self.return_inputs = return_inputs
        self.temperature = temperature
        self.use_inputs_schema = use_inputs_schema
        self.use_outputs_schema = use_outputs_schema
        if schema and streaming:
            streaming = False
        self.streaming = streaming

        predictions = [
            Prediction(
                inputs=example[0],
                outputs=example[1],
                reward=None,
            ).get_json()
            for example in examples
        ]

        if not seed_instructions:
            seed_instructions = []
        self.seed_instructions = seed_instructions

        seed_candidates = [
            {
                "instructions": seed_instruction,
            }
            for seed_instruction in self.seed_instructions
        ]

        self.state = self.add_variable(
            initializer=Instructions(
                instructions=instructions,
                examples=predictions,
                seed_candidates=seed_candidates,
            ).get_json(),
            data_model=Instructions,
            name="state_" + self.name,
        )

    async def call(self, inputs, training=False):
        if not inputs:
            return None
        msgs = self.format_messages(inputs)
        if self.streaming and not training:
            streaming = True
        else:
            streaming = False
        result = await ops.predict(
            msgs,
            schema=self.schema,
            language_model=self.language_model,
            streaming=streaming,
            name="prediction_" + self.name,
            temperature=self.temperature,
        )
        if streaming:
            return result
        if result:
            if training:
                predictions = self.state.get("current_predictions")
                predictions.append(
                    {
                        "inputs": inputs.get_json(),
                        "outputs": result.get_json(),
                        "reward": None,
                    }
                )
            if self.return_inputs:
                return await ops.concat(
                    inputs,
                    result,
                    name="with_inputs_" + self.name,
                )
            else:
                return result
        return None

    async def compute_output_spec(self, inputs, training=False):
        if self.schema:
            if self.return_inputs:
                return await ops.concat(
                    inputs,
                    SymbolicDataModel(
                        schema=self.schema,
                        name=self.name,
                    ),
                    name="with_inputs_" + self.name,
                )
            else:
                return SymbolicDataModel(
                    schema=self.schema,
                    name=self.name,
                )
        else:
            if self.return_inputs:
                return await ops.concat(
                    inputs,
                    SymbolicDataModel(
                        schema=ChatMessage.get_schema(),
                        name=self.name,
                    ),
                    name="with_inputs_" + self.name,
                )
            else:
                return SymbolicDataModel(
                    schema=ChatMessage.get_schema(),
                    name=self.name,
                )

    def format_messages(self, inputs=None):
        template = jinja2.Template(self.prompt_template)
        rendered_prompt = template.render(
            inputs_schema=inputs.get_schema() if self.use_inputs_schema else None,
            outputs_schema=self.schema if self.use_outputs_schema else None,
            examples=[
                (pred.get("inputs"), pred.get("outputs"))
                for pred in self.state.get("examples")
            ],
            instructions=self.state.get("instructions"),
        )
        system_message = ChatMessage(role="system", content=rendered_prompt)

        if inputs and is_chat_messages(inputs):
            msgs = ChatMessages(messages=[system_message] + inputs.get("messages"))
        else:
            user_message = ChatMessage(role="user", content=f"## Input:\n{inputs.get_json()}\n##Output:\n")
            msgs = ChatMessages(messages=[system_message, user_message])
        return msgs

    def get_config(self):
        config = {
            "schema": self.schema,
            "prompt_template": self.prompt_template,
            "examples": self.examples,
            "instructions": self.instructions,
            "seed_instructions": self.seed_instructions,
            "use_inputs_schema": self.use_inputs_schema,
            "use_outputs_schema": self.use_outputs_schema,
            "return_inputs": self.return_inputs,
            "temperature": self.temperature,
            "name": self.name,
            "description": self.description,
            "trainable": self.trainable,
        }
        language_model_config = {
            "language_model": serialization_lib.serialize_synalinks_object(
                self.language_model,
            )
        }
        return {
            **config,
            **language_model_config,
        }

    @classmethod
    def from_config(cls, config):
        language_model = serialization_lib.deserialize_synalinks_object(
            config.pop("language_model"),
        )
        return cls(
            language_model=language_model,
            **config,
        )



# License Apache 2.0: (c) 2025 Yoan Sallami (Synalinks Team)

import asyncio
import logging
import uuid
from typing import Dict, Literal

import requests

from synalinks.src.api_export import synalinks_export
from synalinks.src.backend import api_base
from synalinks.src.backend import api_key
from synalinks.src.callbacks.callback import Callback
from synalinks.src.backend import DataModel


@synalinks_export("synalinks.callbacks.monitor.LogEntry")
class LogEntry(DataModel):
    experiment_id: str
    program_name: str
    program_description: str
    event: Literal["train_begin", "train_end", "batch_begin", "batch_end", "epoch_begin", "epoch_end", "predict_begin", "predict_end"]
    phase: Literal["train", "test", "predict"]
    logs: Dict[str, float]


@synalinks_export("synalinks.callbacks.Monitor")
class Monitor(Callback):
    """Monitor callback for sending training/evaluation/prediction logs to a remote endpoint in realtime.

    This callback sends trace data immediately to a specified endpoint for realtime monitoring
    of training progress, and evaluation metrics.
    Traces are sent asynchronously using asyncio to avoid blocking program execution.

    Args:
        timeout: Request timeout in seconds (default: 5)
        headers: Optional additional headers to include in requests
    """

    def __init__(
        self,
        timeout=5,
        headers=None,
        send_batch_events=False,
        send_predict_events=False,
        send_epoch_events=True,
    ):
        super().__init__()
        self.endpoint = api_base()
        self.timeout = timeout
        if api_key() is not None and not headers:
            headers = {"Authorization": api_key()}
        self.headers = headers or {}
        self._pending_tasks = []
        self.logger = logging.getLogger(__name__)
        self.send_batch_events = send_batch_events
        self.send_predict_events = send_predict_events
        self.send_epoch_events = send_epoch_events
        self._experiment_id = None

    async def _post_trace(self, data: dict):
        """POST trace data to the endpoint asynchronously."""
        url = f"{self.endpoint}/logs"

        try:
            loop = asyncio.get_event_loop()
            # Run requests in executor to make it non-blocking
            response = await loop.run_in_executor(
                None,
                lambda: requests.post(
                    url,
                    json=data,
                    headers=self.headers,
                    timeout=self.timeout,
                )
            )
            response.raise_for_status()
            self.logger.debug(
                f"Trace sent successfully: {data.get('event')} for {data.get('phase')}"
            )
        except requests.exceptions.Timeout:
            self.logger.error(f"Timeout sending trace to {url}")
        except requests.exceptions.RequestException as e:
            self.logger.error(f"Failed to send trace to {url}: {e}")
        except Exception as e:
            self.logger.error(f"Unexpected error sending trace: {e}")

    def _send_trace_async(self, log_data: dict):
        """Send trace asynchronously without blocking."""
        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            # No event loop in current thread, create a new one
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

        # Create task and store reference to prevent garbage collection
        task = loop.create_task(self._post_trace(log_data))
        self._pending_tasks.append(task)

        # Clean up completed tasks
        self._pending_tasks = [t for t in self._pending_tasks if not t.done()]

    def _create_log(self, event: str, phase: str, logs: dict = None):
        """Create a LogEntry and return its JSON representation."""
        log_entry = LogEntry(
            event=event,
            phase=phase,
            experiment_id=self._experiment_id,
            program_name=self.program.name,
            program_description=self.program.description,
            logs=logs or {},
        )
        return log_entry.get_json()

    def on_train_begin(self, logs=None):
        """Called at the beginning of training."""
        self._experiment_id = str(uuid.uuid4())
        log_data = self._create_log(event="train_begin", phase="train", logs=logs)
        self._send_trace_async(log_data)

    def on_train_end(self, logs=None):
        """Called at the end of training."""
        log_data = self._create_log(event="train_end", phase="train", logs=logs)
        self._send_trace_async(log_data)

    def on_epoch_begin(self, epoch, logs=None):
        """Called at the start of an epoch."""
        if not self.send_epoch_events:
            return

        log_data = self._create_log(event="epoch_begin", phase="train", logs=logs)
        self._send_trace_async(log_data)

    def on_epoch_end(self, epoch, logs=None):
        """Called at the end of an epoch."""
        if not self.send_epoch_events:
            return

        log_data = self._create_log(event="epoch_end", phase="train", logs=logs)
        self._send_trace_async(log_data)

    def on_train_batch_begin(self, batch, logs=None):
        """Called at the beginning of a training batch."""
        if not self.send_batch_events:
            return

        log_data = self._create_log(event="batch_begin", phase="train", logs=logs)
        self._send_trace_async(log_data)

    def on_train_batch_end(self, batch, logs=None):
        """Called at the end of a training batch."""
        if not self.send_batch_events:
            return

        log_data = self._create_log(event="batch_end", phase="train", logs=logs)
        self._send_trace_async(log_data)

    def on_test_begin(self, logs=None):
        """Called at the beginning of evaluation or validation."""
        if self._experiment_id is None:
            self._experiment_id = f"test_{uuid.uuid4()}"

        log_data = self._create_log(event="train_begin", phase="test", logs=logs)
        self._send_trace_async(log_data)

    def on_test_end(self, logs=None):
        """Called at the end of evaluation or validation."""
        log_data = self._create_log(event="train_end", phase="test", logs=logs)
        self._send_trace_async(log_data)

    def on_test_batch_begin(self, batch, logs=None):
        """Called at the beginning of a test batch."""
        if not self.send_batch_events:
            return

        log_data = self._create_log(event="batch_begin", phase="test", logs=logs)
        self._send_trace_async(log_data)

    def on_test_batch_end(self, batch, logs=None):
        """Called at the end of a test batch."""
        if not self.send_batch_events:
            return

        log_data = self._create_log(event="batch_end", phase="test", logs=logs)
        self._send_trace_async(log_data)

    def on_predict_begin(self, logs=None):
        """Called at the beginning of prediction."""
        if not self.send_predict_events:
            return

        self._experiment_id = f"predict_{uuid.uuid4()}"
        log_data = self._create_log(event="predict_begin", phase="predict", logs=logs)
        self._send_trace_async(log_data)

    def on_predict_end(self, logs=None):
        """Called at the end of prediction."""
        if not self.send_predict_events:
            return

        log_data = self._create_log(event="predict_end", phase="predict", logs=logs)
        self._send_trace_async(log_data)

    def on_predict_batch_begin(self, batch, logs=None):
        """Called at the beginning of a prediction batch."""
        if not self.send_batch_events:
            return

        log_data = self._create_log(event="batch_begin", phase="predict", logs=logs)
        self._send_trace_async(log_data)

    def on_predict_batch_end(self, batch, logs=None):
        """Called at the end of a prediction batch."""
        if not self.send_batch_events:
            return

        log_data = self._create_log(event="batch_end", phase="predict", logs=logs)
        self._send_trace_async(log_data)

    async def _cleanup(self):
        """Wait for pending tasks."""
        if self._pending_tasks:
            await asyncio.gather(*self._pending_tasks, return_exceptions=True)

    def __del__(self):
        """Cleanup pending traces."""
        if hasattr(self, "_pending_tasks") and self._pending_tasks:
            try:
                loop = asyncio.get_event_loop()
                if loop.is_running():
                    # If loop is running, schedule cleanup
                    loop.create_task(self._cleanup())
                else:
                    # If loop is not running, run cleanup
                    loop.run_until_complete(self._cleanup())
            except Exception:
                pass